from zrb.input.bool_input import BoolInput
from zrb.input.option_input import OptionInput

package_manager_input = OptionInput(
    name="package-manager",
    description="Your package manager",
    prompt="Your package manager",
    options=["apt", "dnf", "pacman", "zypper", "pkg", "brew", "spack"],
    default="apt",
)

use_sudo_input = BoolInput(
    name="use-sudo",
    description="Use sudo or not",
    prompt="Need sudo",
    default=True,
)

setup_bash_input = BoolInput(
    name="setup-bash",
    description="Setup bash",
    prompt="Setup bash",
    default=True,
)

setup_zsh_input = BoolInput(
    name="setup-zsh", description="Setup zsh", prompt="Setup zsh", default=True
)

setup_powershell_input = BoolInput(
    name="setup-powershell",
    description="Setup powershell",
    prompt="Setup powershell",
    default=False,
)



import datetime

import ulid
from pydantic import BaseModel
from sqlmodel import Field, SQLModel


class RoleBase(SQLModel):
    name: str
    description: str = ""


class RoleCreate(RoleBase):
    def with_audit(self, created_by: str) -> "RoleCreateWithAudit":
        return RoleCreateWithAudit(**self.model_dump(), created_by=created_by)


class RoleCreateWithAudit(RoleCreate):
    created_by: str


class RoleCreateWithPermissions(RoleCreate):
    permission_names: list[str] | None = None

    def with_audit(self, created_by: str) -> "RoleCreateWithPermissionsAndAudit":
        return RoleCreateWithPermissionsAndAudit(
            **self.model_dump(), created_by=created_by
        )


class RoleCreateWithPermissionsAndAudit(RoleCreateWithPermissions):
    created_by: str

    def get_role_create_with_audit(self) -> RoleCreateWithAudit:
        data = {
            key: val
            for key, val in self.model_dump().items()
            if key != "permission_names"
        }
        return RoleCreateWithAudit(**data)

    def get_permission_names(self) -> list[str]:
        if self.permission_names is None:
            return []
        return self.permission_names


class RoleUpdate(SQLModel):
    name: str | None = None
    description: str | None = None

    def with_audit(self, updated_by: str) -> "RoleUpdateWithAudit":
        return RoleUpdateWithAudit(
            **self.model_dump(exclude_none=True), updated_by=updated_by
        )


class RoleUpdateWithAudit(RoleUpdate):
    updated_by: str


class RoleUpdateWithPermissions(RoleUpdate):
    permission_names: list[str] | None = None

    def with_audit(self, updated_by: str) -> "RoleUpdateWithPermissionsAndAudit":
        return RoleUpdateWithPermissionsAndAudit(
            **self.model_dump(exclude_none=True), updated_by=updated_by
        )


class RoleUpdateWithPermissionsAndAudit(RoleUpdateWithPermissions):
    updated_by: str

    def get_role_update_with_audit(self) -> RoleUpdateWithAudit:
        data = {
            key: val
            for key, val in self.model_dump(exclude_none=True).items()
            if key != "permission_names"
        }
        return RoleUpdateWithAudit(**data)

    def get_permission_names(self) -> list[str]:
        if self.permission_names is None:
            return []
        return self.permission_names


class RoleResponse(RoleBase):
    id: str
    permission_names: list[str]


class MultipleRoleResponse(BaseModel):
    data: list[RoleResponse]
    count: int


class Role(SQLModel, table=True):
    __tablename__ = "roles"
    id: str = Field(default_factory=lambda: ulid.new().str, primary_key=True)
    created_at: datetime.datetime | None = Field(index=True)
    created_by: str | None = Field(index=True)
    updated_at: datetime.datetime | None = Field(index=True)
    updated_by: str | None = Field(index=True)
    name: str = Field(index=True, unique=True)
    description: str


class RolePermission(SQLModel, table=True):
    __tablename__ = "role_permissions"
    id: str = Field(default_factory=lambda: ulid.new().str, primary_key=True)
    role_id: str = Field(index=True)
    permission_id: str = Field(index=True)
    created_at: datetime.datetime | None
    created_by: str | None



from zrb.builtin.group import (
    uuid_group,
    uuid_v1_group,
    uuid_v3_group,
    uuid_v4_group,
    uuid_v5_group,
)
from zrb.context.any_context import AnyContext
from zrb.input.option_input import OptionInput
from zrb.input.str_input import StrInput
from zrb.task.make_task import make_task


@make_task(
    name="generate-uuid-v1",
    description="🔨 Generate UUID V1",
    input=[
        StrInput(
            name="node",
            description="48-bit hardware address as integer (leave empty for random)",
            prompt="Enter node address (48-bit integer)",
            allow_empty=True,
        ),
        StrInput(
            name="clock-seq",
            description="14-bit sequence number as integer (leave empty for random)",
            prompt="Enter clock sequence (14-bit integer)",
            allow_empty=True,
        ),
    ],
    group=uuid_v1_group,
    alias="generate",
)
def generate_uuid_v1(ctx: AnyContext) -> str:
    import uuid

    result = str(
        uuid.uuid1(
            node=int(ctx.input.node) if ctx.input.node != "" else None,
            clock_seq=int(ctx.input.clock_seq) if ctx.input.clock_seq != "" else None,
        )
    )
    ctx.print(result)
    return result


@make_task(
    name="generate-uuid-v3",
    description="🔨 Generate UUID v3 (MD5 namespace-based)",
    input=[
        OptionInput(
            name="namespace",
            description="Namespace for v3",
            prompt="Select a namespace",
            options=["dns", "url", "oid", "x500"],
        ),
        StrInput(
            name="name",
            description="Name string",
            prompt="Enter the name to namespace-hash",
        ),
    ],
    group=uuid_v3_group,
    alias="generate",
)
def generate_uuid_v3(ctx: AnyContext) -> str:
    import uuid

    ns_map = {
        "dns": uuid.NAMESPACE_DNS,
        "url": uuid.NAMESPACE_URL,
        "oid": uuid.NAMESPACE_OID,
        "x500": uuid.NAMESPACE_X500,
    }
    namespace = ns_map[ctx.input.namespace]
    result = str(uuid.uuid3(namespace, ctx.input.name))
    ctx.print(result)
    return result


@make_task(
    name="generate-uuid-v4",
    description="🔨 Generate UUID v4 (random)",
    group=uuid_v4_group,
    alias="generate",
)
def generate_uuid_v4(ctx: AnyContext) -> str:
    import uuid

    result = str(uuid.uuid4())
    ctx.print(result)
    return result


uuid_group.add_task(generate_uuid_v4, alias="generate")


@make_task(
    name="generate-uuid-v5",
    description="🔨 Generate UUID v5 (SHA1 namespace-based)",
    input=[
        OptionInput(
            name="namespace",
            description="Namespace for v5",
            prompt="Select a namespace",
            options=["dns", "url", "oid", "x500"],
        ),
        StrInput(
            name="name",
            description="Name string",
            prompt="Enter the name to namespace-hash",
        ),
    ],
    group=uuid_v5_group,
    alias="generate",
)
def generate_uuid_v5(ctx: AnyContext) -> str:
    import uuid

    ns_map = {
        "dns": uuid.NAMESPACE_DNS,
        "url": uuid.NAMESPACE_URL,
        "oid": uuid.NAMESPACE_OID,
        "x500": uuid.NAMESPACE_X500,
    }
    namespace = ns_map[ctx.input.namespace]
    result = str(uuid.uuid5(namespace, ctx.input.name))
    ctx.print(result)
    return result


@make_task(
    name="validate-uuid",
    description="✅ Validate UUID",
    input=StrInput(name="id"),
    group=uuid_group,
    alias="validate",
)
def validate_uuid(ctx: AnyContext) -> bool:
    import uuid

    try:
        uuid.UUID(ctx.input.id, version=1)
        ctx.print("Valid UUID")
        return True
    except Exception:
        ctx.print("Invalid UUID")
        return False


@make_task(
    name="validate-uuid-v1",
    description="✅ Validate UUID V1",
    input=StrInput(name="id"),
    group=uuid_v1_group,
    alias="validate",
)
def validate_uuid_v1(ctx: AnyContext) -> bool:
    import uuid

    try:
        uuid.UUID(ctx.input.id, version=1)
        ctx.print("Valid UUID V1")
        return True
    except Exception:
        ctx.print("Invalid UUID V1")
        return False


@make_task(
    name="validate-uuid-v3",
    description="✅ Validate UUID V3",
    input=StrInput(name="id"),
    group=uuid_v3_group,
    alias="validate",
)
def validate_uuid_v3(ctx: AnyContext) -> bool:
    import uuid

    try:
        uuid.UUID(ctx.input.id, version=3)
        ctx.print("Valid UUID V3")
        return True
    except Exception:
        ctx.print("Invalid UUID V3")
        return False


@make_task(
    name="validate-uuid-v4",
    description="✅ Validate UUID V4",
    input=StrInput(name="id"),
    group=uuid_v4_group,
    alias="validate",
)
def validate_uuid_v4(ctx: AnyContext) -> bool:
    import uuid

    try:
        uuid.UUID(ctx.input.id, version=4)
        ctx.print("Valid UUID V4")
        return True
    except Exception:
        ctx.print("Invalid UUID V4")
        return False


@make_task(
    name="validate-uuid-v5",
    description="✅ Validate UUID V5",
    input=StrInput(name="id"),
    group=uuid_v5_group,
    alias="validate",
)
def validate_uuid_v5(ctx: AnyContext) -> bool:
    import uuid

    try:
        uuid.UUID(ctx.input.id, version=5)
        ctx.print("Valid UUID V5")
        return True
    except Exception:
        ctx.print("Invalid UUID V5")
        return False



import os

from my_app_name._zrb.config import APP_DIR
from my_app_name._zrb.util import get_existing_module_names

from zrb.context.any_context import AnyContext
from zrb.util.codemod.modify_dict import append_key_to_dict
from zrb.util.codemod.modify_function import append_code_to_function
from zrb.util.file import read_file, write_file
from zrb.util.string.conversion import (
    to_human_case,
    to_kebab_case,
    to_pascal_case,
    to_snake_case,
)


def is_app_config_file(ctx: AnyContext, file_path: str) -> bool:
    return file_path == os.path.abspath(os.path.join(APP_DIR, "config.py"))


def is_app_main_file(ctx: AnyContext, file_path: str) -> bool:
    return file_path == os.path.abspath(os.path.join(APP_DIR, "main.py"))


def is_gateway_route_file(ctx: AnyContext, file_path: str) -> bool:
    return file_path == os.path.abspath(
        os.path.join(APP_DIR, "module", "gateway", "route.py")
    )


def is_gateway_navigation_config_file(ctx: AnyContext, file_path: str) -> bool:
    return file_path == os.path.abspath(
        os.path.join(APP_DIR, "module", "gateway", "config", "navigation.py")
    )


def is_app_zrb_task_file(ctx: AnyContext, file_path: str) -> bool:
    return file_path == os.path.abspath(os.path.join(APP_DIR, "_zrb", "task.py"))


def is_app_zrb_config_file(ctx: AnyContext, file_path: str) -> bool:
    return file_path == os.path.abspath(os.path.join(APP_DIR, "_zrb", "config.py"))


def is_in_module_dir(ctx: AnyContext, file_path: str) -> bool:
    return file_path.startswith(
        os.path.abspath(
            os.path.join(APP_DIR, "module", to_snake_case(ctx.input.module))
        )
    )


def is_gateway_module_subroute_file(ctx: AnyContext, file_path: str) -> bool:
    module_subroute_file_name = f"{to_snake_case(ctx.input.module)}.py"
    return file_path == os.path.abspath(
        os.path.join(
            APP_DIR, "module", "gateway", "subroute", module_subroute_file_name
        )
    )


def update_app_zrb_config_file(ctx: AnyContext, zrb_config_file_path: str):
    existing_zrb_config_code = read_file(zrb_config_file_path)
    module_name = ctx.input.module
    upper_snake_module_name = to_snake_case(module_name).upper()
    existing_module_names = get_existing_module_names()
    module_port = 3001 + len(
        [
            existing_module_name
            for existing_module_name in existing_module_names
            if existing_module_name != to_snake_case(module_name)
        ]
    )
    write_file(
        file_path=zrb_config_file_path,
        content=append_key_to_dict(
            original_code=existing_zrb_config_code,
            dictionary_name="MICROSERVICES_ENV_VARS",
            new_key=f"MY_APP_NAME_{upper_snake_module_name}_BASE_URL",
            new_value=f"http://localhost:{module_port}",
        ),
    )


def update_app_zrb_task_file(ctx: AnyContext, zrb_task_file_path: str):
    existing_zrb_task_code = read_file(zrb_task_file_path)
    write_file(
        file_path=zrb_task_file_path,
        content=[
            existing_zrb_task_code.strip(),
            "",
            _get_task_definition_code(existing_zrb_task_code, ctx.input.module),
        ],
    )


def _get_task_definition_code(existing_code: str, module_name: str) -> str | None:
    existing_module_names = get_existing_module_names()
    module_port = 3001 + len(
        [
            existing_module_name
            for existing_module_name in existing_module_names
            if existing_module_name != to_snake_case(module_name)
        ]
    )
    snake_module_name = to_snake_case(module_name)
    kebab_module_name = to_kebab_case(module_name)
    pascal_module_name = to_pascal_case(module_name)
    task_definition_code = read_file(
        file_path=os.path.join(
            os.path.dirname(__file__), "template", "module_task_definition.py"
        ),
        replace_map={
            "my_module": snake_module_name,
            "my-module": kebab_module_name,
            "My Module": pascal_module_name,
            "3000": f"{module_port}",
        },
    ).strip()
    if task_definition_code in existing_code:
        return None
    return task_definition_code


def update_app_main_file(ctx: AnyContext, app_main_file_path: str):
    existing_app_main_code = read_file(app_main_file_path)
    write_file(
        file_path=app_main_file_path,
        content=[
            _get_import_module_route_code(existing_app_main_code, ctx.input.module),
            existing_app_main_code,
            "",
            _get_assert_module_route_code(existing_app_main_code, ctx.input.module),
        ],
    )


def _get_import_module_route_code(existing_code: str, module_name: str) -> str | None:
    snake_module_name = to_snake_case(module_name)
    import_module_path = f"my_app_name.module.{snake_module_name}"
    import_route_code = (
        f"from {import_module_path} import route as {snake_module_name}_route"
    )
    if import_route_code in existing_code:
        return None
    return import_route_code


def _get_assert_module_route_code(existing_code: str, module_name: str) -> str | None:
    snake_module_name = to_snake_case(module_name)
    assert_route_code = f"assert {snake_module_name}_route"
    return assert_route_code if assert_route_code not in existing_code else None


def update_app_config_file(ctx: AnyContext, module_config_path: str):
    existing_config_code = read_file(module_config_path)
    write_file(
        module_config_path,
        [
            existing_config_code.strip(),
            _get_new_module_config_code(existing_config_code, ctx.input.module),
        ],
    )


def _get_new_module_config_code(existing_code: str, module_name: str) -> str | None:
    existing_module_names = get_existing_module_names()
    module_port = 3000 + len(
        [
            existing_module_name
            for existing_module_name in existing_module_names
            if existing_module_name != to_snake_case(module_name)
        ]
    )
    module_base_url = f"http://localhost:{module_port}"
    upper_snake_module_name = to_snake_case(module_name).upper()
    config_name = f"APP_{upper_snake_module_name}_BASE_URL"
    env_name = f"MY_APP_NAME_{upper_snake_module_name}_BASE_URL"
    config_code = f'{config_name} = os.getenv("{env_name}", "{module_base_url}")'
    if config_code in existing_code:
        return None
    return config_code


def update_gateway_route_file(ctx: AnyContext, gateway_route_file_path: str):
    existing_gateway_route_code = read_file(gateway_route_file_path)
    snake_module_name = to_snake_case(ctx.input.module)
    write_file(
        file_path=gateway_route_file_path,
        content=[
            _get_module_subroute_import(existing_gateway_route_code, ctx.input.module),
            append_code_to_function(
                original_code=existing_gateway_route_code,
                function_name="serve_route",
                new_code="\n".join(
                    [
                        f"# Serve {snake_module_name} route",
                        f"serve_{snake_module_name}_route(app)",
                    ]
                ),
            ),
        ],
    )


def _get_module_subroute_import(existing_code: str, module_name: str) -> str | None:
    snake_module_name = to_snake_case(module_name)
    import_module_path = f"my_app_name.module.gateway.subroute.{snake_module_name}"
    import_code = f"from {import_module_path} import serve_{snake_module_name}_route"
    if import_code in existing_code:
        return None
    return import_code


def update_gateway_navigation_config_file(
    ctx: AnyContext, gateway_navigation_config_file_path: str
):
    existing_gateway_navigation_config_code = read_file(
        gateway_navigation_config_file_path
    )
    snake_module_name = to_snake_case(ctx.input.module)
    kebab_module_name = to_kebab_case(ctx.input.module)
    human_module_name = to_human_case(ctx.input.module)
    new_navigation_config_code = read_file(
        file_path=os.path.join(
            os.path.dirname(__file__), "template", "navigation_config_file.py"
        ),
        replace_map={
            "my_module": snake_module_name,
            "my-module": kebab_module_name,
            "My Module": human_module_name.title(),
        },
    ).strip()
    write_file(
        file_path=gateway_navigation_config_file_path,
        content=[
            existing_gateway_navigation_config_code,
            new_navigation_config_code,
        ],
    )



import os

from my_app_name._zrb.config import APP_DIR

from zrb.context.any_context import AnyContext
from zrb.util.codemod.modify_class import append_code_to_class
from zrb.util.codemod.modify_class_parent import prepend_parent_class
from zrb.util.codemod.modify_function import append_code_to_function
from zrb.util.codemod.modify_module import prepend_code_to_module
from zrb.util.file import read_file, write_file
from zrb.util.string.conversion import (
    to_human_case,
    to_kebab_case,
    to_pascal_case,
    to_snake_case,
)


def is_gateway_navigation_config_file(ctx: AnyContext, file_path: str) -> bool:
    return file_path == os.path.abspath(
        os.path.join(APP_DIR, "module", "gateway", "config", "navigation.py")
    )


def get_add_permission_migration_script(ctx: AnyContext) -> str:
    kebab_entity_name = to_kebab_case(ctx.input.entity)
    return "\n".join(
        [
            "op.bulk_insert(",
            '    metadata.tables["permissions"],',
            "    [",
            f'        {{"name": "{kebab_entity_name}:create", "description": "create {kebab_entity_name}"}},',  # noqa
            f'        {{"name": "{kebab_entity_name}:read", "description": "read {kebab_entity_name}"}},',  # noqa
            f'        {{"name": "{kebab_entity_name}:update", "description": "update {kebab_entity_name}"}},',  # noqa
            f'        {{"name": "{kebab_entity_name}:delete", "description": "delete {kebab_entity_name}"}},',  # noqa
            "    ]",
            ")",
        ]
    )


def get_remove_permission_migration_script(ctx: AnyContext) -> str:
    kebab_entity_name = to_kebab_case(ctx.input.entity)
    return "\n".join(
        [
            "op.execute(",
            '    sa.delete(metadata.tables["permissions"])',
            '    .where(metadata.tables["permissions"].c.name.in_(',
            f'        "{kebab_entity_name}:create",',
            f'        "{kebab_entity_name}:read",',
            f'        "{kebab_entity_name}:update",',
            f'        "{kebab_entity_name}:delete",',
            "    ))",
            ")",
        ]
    )


def get_auth_migration_version_dir() -> str:
    return os.path.join(APP_DIR, "module", "auth", "migration", "versions")


def get_existing_auth_migration_file_names() -> list[str]:
    migration_version_dir = get_auth_migration_version_dir()
    return [
        file_name
        for file_name in os.listdir(migration_version_dir)
        if file_name.endswith(".py")
    ]


def get_existing_auth_migration_xcom_key(ctx: AnyContext) -> str:
    snake_entity_name = to_snake_case(ctx.input.entity)
    return f"existing_my_app_name_auth_{snake_entity_name}_migration"


def is_in_app_schema_dir(ctx: AnyContext, file_path: str) -> bool:
    return file_path.startswith(
        os.path.abspath(
            os.path.join(APP_DIR, "schema", to_snake_case(ctx.input.entity))
        )
    )


def is_in_module_entity_test_dir(ctx: AnyContext, file_path: str) -> bool:
    return file_path.startswith(
        os.path.abspath(
            os.path.join(
                APP_DIR,
                "test",
                to_snake_case(ctx.input.module),
                to_snake_case(ctx.input.entity),
            )
        )
    )


def is_in_module_entity_dir(ctx: AnyContext, file_path: str) -> bool:
    return file_path.startswith(
        os.path.abspath(
            os.path.join(
                APP_DIR,
                "module",
                to_snake_case(ctx.input.module),
                "service",
                to_snake_case(ctx.input.entity),
            )
        )
    )


def is_module_route_file(ctx: AnyContext, file_path: str) -> bool:
    module_route_file = os.path.abspath(
        os.path.join(
            APP_DIR,
            "module",
            to_snake_case(ctx.input.module),
            "route.py",
        )
    )
    return file_path == module_route_file


def is_module_migration_metadata_file(ctx: AnyContext, file_path: str) -> bool:
    module_migration_metadata_file = os.path.abspath(
        os.path.join(
            APP_DIR,
            "module",
            to_snake_case(ctx.input.module),
            "migration_metadata.py",
        )
    )
    return file_path == module_migration_metadata_file


def is_module_client_file(ctx: AnyContext, file_path: str) -> bool:
    module_any_client_file = os.path.abspath(
        os.path.join(
            APP_DIR,
            "module",
            to_snake_case(ctx.input.module),
            "client",
            f"{to_snake_case(ctx.input.module)}_client.py",
        )
    )
    return file_path == module_any_client_file


def is_module_api_client_file(ctx: AnyContext, file_path: str) -> bool:
    module_api_client_file = os.path.abspath(
        os.path.join(
            APP_DIR,
            "module",
            to_snake_case(ctx.input.module),
            "client",
            f"{to_snake_case(ctx.input.module)}_api_client.py",
        )
    )
    return file_path == module_api_client_file


def is_module_direct_client_file(ctx: AnyContext, file_path: str) -> bool:
    module_direct_client_file = os.path.abspath(
        os.path.join(
            APP_DIR,
            "module",
            to_snake_case(ctx.input.module),
            "client",
            f"{to_snake_case(ctx.input.module)}_direct_client.py",
        )
    )
    return file_path == module_direct_client_file


def is_module_gateway_subroute_file(ctx: AnyContext, file_path: str) -> bool:
    module_gateway_subroute_file = os.path.abspath(
        os.path.join(
            APP_DIR,
            "module",
            "gateway",
            "subroute",
            f"{to_snake_case(ctx.input.module)}.py",
        )
    )
    return file_path == module_gateway_subroute_file


def is_module_gateway_subroute_view_file(ctx: AnyContext, file_path: str) -> bool:
    module_gateway_subroute_file = os.path.abspath(
        os.path.join(
            APP_DIR,
            "module",
            "gateway",
            "view",
            "content",
            f"{to_kebab_case(ctx.input.module)}",
            f"{to_kebab_case(ctx.input.entity)}.html",
        )
    )
    return file_path == module_gateway_subroute_file


def update_migration_metadata_file(ctx: AnyContext, migration_metadata_file_path: str):
    app_name = os.path.basename(APP_DIR)
    existing_migration_metadata_code = read_file(migration_metadata_file_path)
    write_file(
        file_path=migration_metadata_file_path,
        content=[
            _get_migration_import_schema_code(
                existing_migration_metadata_code, app_name, ctx.input.entity
            ),
            existing_migration_metadata_code.strip(),
            _get_migration_entity_metadata_assignment_code(
                existing_migration_metadata_code, ctx.input.entity
            ),
        ],
    )


def _get_migration_import_schema_code(
    existing_code: str, app_name: str, entity_name: str
) -> str | None:
    snake_entity_name = to_snake_case(entity_name)
    pascal_entity_name = to_pascal_case(entity_name)
    import_module_path = f"{app_name}.schema.{snake_entity_name}"
    import_schema_code = f"from {import_module_path} import {pascal_entity_name}"
    if import_schema_code in existing_code:
        return None
    return import_schema_code


def _get_migration_entity_metadata_assignment_code(
    existing_code: str, entity_name: str
) -> str | None:
    pascal_entity_name = to_pascal_case(entity_name)
    entity_metadata_assignment_code = "\n".join(
        [
            f"{pascal_entity_name}.metadata = metadata",
            f"{pascal_entity_name}.__table__.tometadata(metadata)",
        ]
    )
    if entity_metadata_assignment_code in existing_code:
        return None
    return entity_metadata_assignment_code


def update_client_file(ctx: AnyContext, client_file_path: str):
    existing_client_code = read_file(client_file_path)
    pascal_module_name = to_pascal_case(ctx.input.module)
    snake_entity_name = to_snake_case(ctx.input.entity)
    snake_plural_entity_name = to_snake_case(ctx.input.plural)
    pascal_entity_name = to_pascal_case(ctx.input.entity)
    write_file(
        file_path=client_file_path,
        content=[
            _get_import_schema_for_client_code(
                existing_code=existing_client_code, entity_name=ctx.input.entity
            ),
            append_code_to_class(
                original_code=existing_client_code,
                class_name=f"{pascal_module_name}Client",
                new_code=read_file(
                    file_path=os.path.join(
                        os.path.dirname(__file__), "template", "client_method.py"
                    ),
                    replace_map={
                        "my_entity": snake_entity_name,
                        "my_entities": snake_plural_entity_name,
                        "MyEntity": pascal_entity_name,
                    },
                ),
            ),
        ],
    )


def _get_import_schema_for_client_code(
    existing_code: str, entity_name: str
) -> str | None:
    snake_entity_name = to_snake_case(entity_name)
    pascal_entity_name = to_pascal_case(entity_name)
    schema_import_path = f"my_app_name.schema.{snake_entity_name}"
    new_code = "\n".join(
        [
            f"from {schema_import_path} import (",
            f"   Multiple{pascal_entity_name}Response,",
            f"   {pascal_entity_name}Create,",
            f"   {pascal_entity_name}CreateWithAudit,",
            f"   {pascal_entity_name}Response,",
            f"   {pascal_entity_name}Update,",
            f"   {pascal_entity_name}UpdateWithAudit,",
            ")",
        ]
    )
    if new_code in existing_code:
        return None
    return new_code


def update_api_client_file(ctx: AnyContext, api_client_file_path: str):
    existing_api_client_code = read_file(api_client_file_path)
    upper_snake_module_name = to_snake_case(ctx.input.module).upper()
    app_name = os.path.basename(APP_DIR)
    snake_entity_name = to_snake_case(ctx.input.entity)
    snake_module_name = to_snake_case(ctx.input.module)
    pascal_module_name = to_pascal_case(ctx.input.module)
    write_file(
        file_path=api_client_file_path,
        content=[
            f"from {app_name}.module.{snake_module_name}.service.{snake_entity_name}.{snake_entity_name}_service_factory import {snake_entity_name}_service",  # noqa
            prepend_code_to_module(
                prepend_parent_class(
                    original_code=existing_api_client_code,
                    class_name=f"{pascal_module_name}APIClient",
                    parent_class_name=f"{snake_entity_name}_api_client",
                ),
                f"{snake_entity_name}_api_client = {snake_entity_name}_service.as_api_client(base_url=APP_{upper_snake_module_name}_BASE_URL)",  # noqa
            ),
        ],
    )


def update_direct_client_file(ctx: AnyContext, direct_client_file_path: str):
    existing_direct_client_code = read_file(direct_client_file_path)
    app_name = os.path.basename(APP_DIR)
    snake_entity_name = to_snake_case(ctx.input.entity)
    snake_module_name = to_snake_case(ctx.input.module)
    pascal_module_name = to_pascal_case(ctx.input.module)
    write_file(
        file_path=direct_client_file_path,
        content=[
            f"from {app_name}.module.{snake_module_name}.service.{snake_entity_name}.{snake_entity_name}_service_factory import {snake_entity_name}_service",  # noqa
            prepend_code_to_module(
                prepend_parent_class(
                    original_code=existing_direct_client_code,
                    class_name=f"{pascal_module_name}DirectClient",
                    parent_class_name=f"{snake_entity_name}_direct_client",
                ),
                f"{snake_entity_name}_direct_client = {snake_entity_name}_service.as_direct_client()",  # noqa
            ).strip(),
        ],
    )


def update_route_file(ctx: AnyContext, route_file_path: str):
    existing_route_code = read_file(route_file_path)
    entity_name = to_snake_case(ctx.input.entity)
    app_name = os.path.basename(APP_DIR)
    module_name = to_snake_case(ctx.input.module)
    write_file(
        file_path=route_file_path,
        content=[
            f"from {app_name}.module.{module_name}.service.{entity_name}.{entity_name}_service_factory import {entity_name}_service",  # noqa
            append_code_to_function(
                existing_route_code,
                "serve_route",
                f"{entity_name}_service.serve_route(app)",
            ),
        ],
    )


def update_gateway_subroute_file(ctx: AnyContext, module_gateway_subroute_path: str):
    snake_module_name = to_snake_case(ctx.input.module)
    kebab_module_name = to_kebab_case(ctx.input.module)
    snake_entity_name = to_snake_case(ctx.input.entity)
    kebab_entity_name = to_kebab_case(ctx.input.entity)
    snake_plural_entity_name = to_snake_case(ctx.input.plural)
    kebab_plural_entity_name = to_kebab_case(ctx.input.plural)
    pascal_entity_name = to_pascal_case(ctx.input.entity)
    existing_gateway_subroute_code = read_file(module_gateway_subroute_path)
    write_file(
        file_path=module_gateway_subroute_path,
        content=[
            _get_import_client_for_gateway_subroute_code(
                existing_gateway_subroute_code, module_name=ctx.input.module
            ),
            _get_import_schema_for_gateway_subroute_code(
                existing_gateway_subroute_code, entity_name=ctx.input.entity
            ),
            append_code_to_function(
                original_code=existing_gateway_subroute_code,
                function_name=f"serve_{snake_module_name}_route",
                new_code=read_file(
                    file_path=os.path.join(
                        os.path.dirname(__file__), "template", "gateway_subroute.py"
                    ),
                    replace_map={
                        "my_module": snake_module_name,
                        "my-module": kebab_module_name,
                        "my_entity": snake_entity_name,
                        "my-entity": kebab_entity_name,
                        "my_entities": snake_plural_entity_name,
                        "my-entities": kebab_plural_entity_name,
                        "MyEntity": pascal_entity_name,
                    },
                ),
            ),
        ],
    )


def _get_import_client_for_gateway_subroute_code(
    existing_code: str, module_name: str
) -> str | None:
    snake_module_name = to_snake_case(module_name)
    client_import_path = f"my_app_name.module.{snake_module_name}.client.{snake_module_name}_client_factory"  # noqa
    new_code = f"from {client_import_path} import {snake_module_name}_client"
    if new_code in existing_code:
        return None
    return new_code


def _get_import_schema_for_gateway_subroute_code(
    existing_code: str, entity_name: str
) -> str | None:
    snake_entity_name = to_snake_case(entity_name)
    pascal_entity_name = to_pascal_case(entity_name)
    schema_import_path = f"my_app_name.schema.{snake_entity_name}"
    new_code = "\n".join(
        [
            f"from {schema_import_path} import (",
            f"   Multiple{pascal_entity_name}Response,",
            f"   {pascal_entity_name}Create,",
            f"   {pascal_entity_name}Response,",
            f"   {pascal_entity_name}Update,",
            ")",
        ]
    )
    if new_code in existing_code:
        return None
    return new_code


def update_gateway_navigation_config_file(
    ctx: AnyContext, gateway_navigation_config_file_path: str
):
    existing_gateway_navigation_config_code = read_file(
        gateway_navigation_config_file_path
    )
    snake_module_name = to_snake_case(ctx.input.module)
    kebab_module_name = to_kebab_case(ctx.input.module)
    kebab_entity_name = to_kebab_case(ctx.input.entity)
    human_entity_name = to_human_case(ctx.input.entity)
    kebab_plural_name = to_kebab_case(ctx.input.plural)
    new_navigation_config_code = read_file(
        file_path=os.path.join(
            os.path.dirname(__file__), "template", "navigation_config_file.py"
        ),
        replace_map={
            "my_module": snake_module_name,
            "my-module": kebab_module_name,
            "my-entity": kebab_entity_name,
            "My Entity": human_entity_name.title(),
            "my-entities": kebab_plural_name,
        },
    ).strip()
    write_file(
        file_path=gateway_navigation_config_file_path,
        content=[
            existing_gateway_navigation_config_code,
            new_navigation_config_code,
        ],
    )



import json
from collections.abc import Callable
from typing import Any

from zrb.util.cli.style import stylize_faint


async def print_node(
    print_func: Callable, agent_run: Any, node: Any, log_indent_level: int = 0
):
    """Prints the details of an agent execution node using a provided print function."""
    from pydantic_ai import Agent
    from pydantic_ai.exceptions import UnexpectedModelBehavior
    from pydantic_ai.messages import (
        FinalResultEvent,
        FunctionToolCallEvent,
        FunctionToolResultEvent,
        PartDeltaEvent,
        PartStartEvent,
        TextPartDelta,
        ThinkingPartDelta,
        ToolCallPartDelta,
    )

    meta = getattr(node, "id", None) or getattr(node, "request_id", None)
    if Agent.is_user_prompt_node(node):
        print_func(_format_header("🔠 Receiving input...", log_indent_level))
    elif Agent.is_model_request_node(node):
        # A model request node => We can stream tokens from the model's request
        print_func(_format_header("🧠 Processing...", log_indent_level))
        # Reference: https://ai.pydantic.dev/agents/#streaming
        try:
            async with node.stream(agent_run.ctx) as request_stream:
                is_streaming = False
                async for event in request_stream:
                    if isinstance(event, PartStartEvent) and event.part:
                        if is_streaming:
                            print_func("")
                        content = _get_event_part_content(event)
                        print_func(_format_content(content, log_indent_level), end="")
                        is_streaming = True
                    elif isinstance(event, PartDeltaEvent):
                        if isinstance(event.delta, TextPartDelta) or isinstance(
                            event.delta, ThinkingPartDelta
                        ):
                            content_delta = event.delta.content_delta
                            print_func(
                                _format_stream_content(content_delta, log_indent_level),
                                end="",
                            )
                        elif isinstance(event.delta, ThinkingPartDelta):
                            content_delta = event.delta.content_delta
                            print_func(
                                _format_stream_content(content_delta, log_indent_level),
                                end="",
                            )
                        elif isinstance(event.delta, ToolCallPartDelta):
                            args_delta = event.delta.args_delta
                            print_func(
                                _format_stream_content(args_delta, log_indent_level),
                                end="",
                            )
                        is_streaming = True
                    elif isinstance(event, FinalResultEvent) and event.tool_name:
                        if is_streaming:
                            print_func("")
                        tool_name = event.tool_name
                        print_func(
                            _format_content(
                                f"Result: tool_name={tool_name}", log_indent_level
                            )
                        )
                        is_streaming = False
                if is_streaming:
                    print_func("")
        except UnexpectedModelBehavior as e:
            print_func("")  # ensure newline consistency
            print_func(
                _format_content(
                    (
                        f"⚠️ Unexpected Model Behavior: {e}. "
                        f"Cause: {e.__cause__}. Node.Id: {meta}"
                    ),
                    log_indent_level,
                )
            )
    elif Agent.is_call_tools_node(node):
        # A handle-response node => The model returned some data, potentially calls a tool
        print_func(_format_header("🧰 Calling Tool...", log_indent_level))
        try:
            async with node.stream(agent_run.ctx) as handle_stream:
                async for event in handle_stream:
                    if isinstance(event, FunctionToolCallEvent):
                        args = _get_event_part_args(event)
                        call_id = event.part.tool_call_id
                        tool_name = event.part.tool_name
                        print_func(
                            _format_content(
                                f"{call_id} | Call {tool_name} {args}", log_indent_level
                            )
                        )
                    elif isinstance(event, FunctionToolResultEvent):
                        call_id = event.tool_call_id
                        result_content = event.result.content
                        print_func(
                            _format_content(
                                f"{call_id} | {result_content}", log_indent_level
                            )
                        )
        except UnexpectedModelBehavior as e:
            print_func("")  # ensure newline consistency
            print_func(
                _format_content(
                    (
                        f"⚠️ Unexpected Model Behavior: {e}. "
                        f"Cause: {e.__cause__}. Node.Id: {meta}"
                    ),
                    log_indent_level,
                )
            )
    elif Agent.is_end_node(node):
        # Once an End node is reached, the agent run is complete
        print_func(_format_header("✅ Completed...", log_indent_level))


def _format_header(text: str, log_indent_level: int = 0) -> str:
    return _format(
        text,
        base_indent=2,
        first_indent=0,
        indent=0,
        log_indent_level=log_indent_level,
    )


def _format_content(text: str, log_indent_level: int = 0) -> str:
    return _format(
        text,
        base_indent=2,
        first_indent=3,
        indent=3,
        log_indent_level=log_indent_level,
    )


def _format_stream_content(text: str, log_indent_level: int = 0) -> str:
    return _format(
        text,
        base_indent=2,
        indent=3,
        log_indent_level=log_indent_level,
        is_stream=True,
    )


def _format(
    text: str,
    base_indent: int = 0,
    first_indent: int = 0,
    indent: int = 0,
    log_indent_level: int = 0,
    is_stream: bool = False,
) -> str:
    line_prefix = (base_indent * (log_indent_level + 1) + indent) * " "
    processed_text = text.replace("\n", f"\n{line_prefix}")
    if is_stream:
        return stylize_faint(processed_text)
    first_line_prefix = (base_indent * (log_indent_level + 1) + first_indent) * " "
    return stylize_faint(f"{first_line_prefix}{processed_text}")


def _get_event_part_args(event: Any) -> Any:
    # Handle empty arguments across different providers
    if event.part.args == "" or event.part.args is None:
        return {}
    if isinstance(event.part.args, str):
        # Some providers might send "null" or "{}" as a string
        if event.part.args.strip() in ["null", "{}"]:
            return {}
        try:
            obj = json.loads(event.part.args)
            if isinstance(obj, dict):
                return _truncate_kwargs(obj)
        except json.JSONDecodeError:
            pass
    # Handle dummy property if present (from our schema sanitization)
    if isinstance(event.part.args, dict):
        return _truncate_kwargs(event.part.args)
    print(type(event.part.args))
    return event.part.args


def _truncate_kwargs(kwargs: dict[str, Any]) -> dict[str, Any]:
    return {key: _truncate_arg(val) for key, val in kwargs.items() if key != "_dummy"}


def _truncate_arg(arg: str, length: int = 19) -> str:
    if isinstance(arg, str) and len(arg) > length:
        return f"{arg[:length-4]} ..."
    return arg


def _get_event_part_content(event: Any) -> str:
    if not hasattr(event, "part"):
        return f"{event}"
    if not hasattr(event.part, "content"):
        return f"{event.part}"
    return getattr(event.part, "content")



import random

from zrb.builtin.group import random_group
from zrb.context.any_context import AnyContext
from zrb.input.int_input import IntInput
from zrb.input.str_input import StrInput
from zrb.task.make_task import make_task


@make_task(
    name="throw-dice",
    description="🔢 Throw dices",
    input=[
        StrInput(
            name="side",
            description="Number of sides",
            prompt="How many sides (comma separated)",
            default="6",
        ),
        IntInput(
            name="num-rolls",
            description="Number of rolls",
            prompt="How many rolls",
            default="1",
        ),
    ],
    retries=0,
    group=random_group,
    alias="throw",
)
def throw_dice(ctx: AnyContext) -> str:
    dice = [int(side.strip()) for side in ctx.input.side.split(",")]
    num_rolls = ctx.input.num_rolls
    str_sums = []
    for i in range(num_rolls):
        throw_results = [random.randint(1, sides) for sides in dice]
        throw_results_sum = sum(throw_results)
        throw_results_detail = ", ".join(str(num) for num in throw_results)
        ctx.print(f"#{i+1}: {throw_results_detail} -> {throw_results_sum}")
        str_sums.append(str(throw_results_sum))
    return "\n".join(str_sums)


@make_task(
    name="shuffle",
    description="🔀 Shuffle list",
    input=StrInput(
        name="values",
        description="Value to be shuffled",
        prompt="List of values (comma separated)",
        default="🪙, 🪄, ⚔️, 🍷",
    ),
    retries=0,
    group=random_group,
    alias="shuffle",
)
def shuffle_values(ctx: AnyContext) -> str:
    shuffled = [value.strip() for value in ctx.input["values"].split(",")]
    ctx.print("Shuffling...")
    random.shuffle(shuffled)
    return "\n".join(shuffled)



import os
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from zrb.util.cmd.command import run_command

if TYPE_CHECKING:
    from zrb.util.git_diff_model import DiffResult


async def get_diff(
    repo_dir: str,
    source_commit: str,
    current_commit: str,
    print_method: Callable[..., Any] = print,
) -> "DiffResult":
    """
    Get the difference between two commits in a Git repository.

    Args:
        repo_dir (str): The path to the Git repository.
        source_commit (str): The source commit hash or reference.
        current_commit (str): The current commit hash or reference.
        print_method (Callable[..., Any]): Method to print command output.

    Returns:
        DiffResult: An object containing lists of created, removed, and updated files.

    Raises:
        Exception: If the git command returns a non-zero exit code.
    """
    from zrb.util.git_diff_model import DiffResult

    cmd_result, exit_code = await run_command(
        cmd=["git", "diff", source_commit, current_commit],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")
    lines = cmd_result.output.strip().split("\n")
    diff: dict[str, dict[str, bool]] = {}
    for line in lines:
        if not line.startswith("---") and not line.startswith("+++"):
            continue
        if line[4:6] != "a/" and line[4:6] != "b/":
            continue
        # line should contains something like `--- a/some-file.txt`
        file = line[6:]
        if file not in diff:
            diff[file] = {"plus": False, "minus": False}
        if line.startswith("---"):
            diff[file]["minus"] = True
        if line.startswith("+++"):
            diff[file]["plus"] = True
    return DiffResult(
        created=[
            file for file, state in diff.items() if state["plus"] and not state["minus"]
        ],
        removed=[
            file for file, state in diff.items() if not state["plus"] and state["minus"]
        ],
        updated=[
            file for file, state in diff.items() if state["plus"] and state["minus"]
        ],
    )


async def get_repo_dir(print_method: Callable[..., Any] = print) -> str:
    """
    Get the top-level directory of the Git repository.

    Args:
        print_method (Callable[..., Any]): Method to print command output.

    Returns:
        str: The absolute path to the repository's top-level directory.

    Raises:
        Exception: If the git command returns a non-zero exit code.
    """
    cmd_result, exit_code = await run_command(
        cmd=["git", "rev-parse", "--show-toplevel"],
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")
    return os.path.abspath(cmd_result.output.strip())


async def get_current_branch(
    repo_dir: str, print_method: Callable[..., Any] = print
) -> str:
    """
    Get the current branch name of the Git repository.

    Args:
        repo_dir (str): The path to the Git repository.
        print_method (Callable[..., Any]): Method to print command output.

    Returns:
        str: The name of the current branch.

    Raises:
        Exception: If the git command returns a non-zero exit code.
    """
    cmd_result, exit_code = await run_command(
        cmd=["git", "rev-parse", "--abbrev-ref", "HEAD"],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")
    return cmd_result.output.strip()


async def get_branches(
    repo_dir: str, print_method: Callable[..., Any] = print
) -> list[str]:
    """
    Get a list of all branches in the Git repository.

    Args:
        repo_dir (str): The path to the Git repository.
        print_method (Callable[..., Any]): Method to print command output.

    Returns:
        list[str]: A list of branch names.

    Raises:
        Exception: If the git command returns a non-zero exit code.
    """
    cmd_result, exit_code = await run_command(
        cmd=["git", "rev-parse", "--abbrev-ref", "HEAD"],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")
    return [
        branch.lstrip("*").strip() for branch in cmd_result.output.strip().split("\n")
    ]


async def delete_branch(
    repo_dir: str, branch_name: str, print_method: Callable[..., Any] = print
) -> str:
    """
    Delete a branch in the Git repository.

    Args:
        repo_dir (str): The path to the Git repository.
        branch_name (str): The name of the branch to delete.
        print_method (Callable[..., Any]): Method to print command output.

    Returns:
        str: The output of the git command.

    Raises:
        Exception: If the git command returns a non-zero exit code.
    """
    cmd_result, exit_code = await run_command(
        cmd=["git", "branch", "-D", branch_name],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")
    return cmd_result.output.strip()


async def add(repo_dir: str, print_method: Callable[..., Any] = print):
    """
    Add all changes to the Git staging area.

    Args:
        repo_dir (str): The path to the Git repository.
        print_method (Callable[..., Any]): Method to print command output.

    Raises:
        Exception: If the git command returns a non-zero exit code.
    """
    _, exit_code = await run_command(
        cmd=["git", "add", ".", "-A"],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")


async def commit(
    repo_dir: str, message: str, print_method: Callable[..., Any] = print
) -> str:
    """
    Commit changes in the Git repository.

    Args:
        repo_dir (str): The path to the Git repository.
        message (str): The commit message.
        print_method (Callable[..., Any]): Method to print command output.

    Raises:
        Exception: If the git command returns a non-zero exit code, unless it's
            the "nothing to commit" message.
    """
    cmd_result, exit_code = await run_command(
        cmd=["git", "commit", "-m", message],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        ignored_error_message = "nothing to commit, working tree clean"
        if (
            ignored_error_message not in cmd_result.error
            and ignored_error_message not in cmd_result.output
        ):
            raise Exception(f"Non zero exit code: {exit_code}")


async def pull(
    repo_dir: str, remote: str, branch: str, print_method: Callable[..., Any] = print
) -> str:
    """
    Pull changes from a remote repository and branch.

    Args:
        repo_dir (str): The path to the Git repository.
        remote (str): The name of the remote.
        branch (str): The name of the branch.
        print_method (Callable[..., Any]): Method to print command output.

    Raises:
        Exception: If the git command returns a non-zero exit code.
    """
    _, exit_code = await run_command(
        cmd=["git", "pull", remote, branch],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")


async def push(
    repo_dir: str, remote: str, branch: str, print_method: Callable[..., Any] = print
) -> str:
    """
    Push changes to a remote repository and branch.

    Args:
        repo_dir (str): The path to the Git repository.
        remote (str): The name of the remote.
        branch (str): The name of the branch.
        print_method (Callable[..., Any]): Method to print command output.

    Raises:
        Exception: If the git command returns a non-zero exit code.
    """
    _, exit_code = await run_command(
        cmd=["git", "push", "-u", remote, branch],
        cwd=repo_dir,
        print_method=print_method,
    )
    if exit_code != 0:
        raise Exception(f"Non zero exit code: {exit_code}")



import json
from collections.abc import Callable
from urllib.parse import urljoin


async def open_web_page(url: str) -> str:
    """
    Fetches, parses, and converts the content of a web page to Markdown.

    This tool "reads" a web page by fetching its content, stripping away
    non-essential elements like scripts and styles, and then converting the
    cleaned HTML into Markdown format. This preserves the semantic structure
    of the content (headings, lists, etc.) while removing clutter. It also
    extracts all hyperlinks and resolves them to absolute URLs.

    Args:
        url (str): The full URL of the web page to open (e.g.,
            "https://example.com/article").

    Returns:
        str: A JSON object containing the page's content in Markdown format
            and a list of all absolute links found on the page.
    """
    html_content, links = await _fetch_page_content(url)
    markdown_content = _convert_html_to_markdown(html_content)
    return json.dumps({"content": markdown_content, "links_on_page": links})


def create_search_internet_tool(serp_api_key: str) -> Callable[[str, int], str]:
    """
    Creates a tool that searches the internet using the SerpAPI Google Search
    API.

    This factory returns a function that can be used to find information on the
    web. The generated tool is the primary way to answer general knowledge
    questions or to find information on topics you are unfamiliar with.

    Args:
        serp_api_key (str): The API key for SerpAPI.

    Returns:
        Callable: A function that takes a search query and returns a list of
            search results.
    """

    def search_internet(query: str, num_results: int = 10) -> str:
        """
        Performs an internet search using Google and returns a summary of the results.

        Use this tool to find information on the web, answer general knowledge questions, or research topics.

        Args:
            query (str): The search query.
            num_results (int, optional): The desired number of search results. Defaults to 10.

        Returns:
            str: A formatted string summarizing the search results, including titles, links, and snippets.
        """
        import requests

        response = requests.get(
            "https://serpapi.com/search",
            headers={
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
            },
            params={
                "q": query,
                "num": num_results,
                "hl": "en",
                "safe": "off",
                "api_key": serp_api_key,
            },
        )
        if response.status_code != 200:
            raise Exception(
                f"Error: Unable to retrieve search results (status code: {response.status_code})"
            )
        return response.json()

    return search_internet


def search_wikipedia(query: str) -> str:
    """
    Searches for articles on Wikipedia.

    This is a specialized search tool for querying Wikipedia. It's best for
    when the user is asking for definitions, historical information, or
    biographical details that are likely to be found on an encyclopedia.

    Args:
        query (str): The search term or question.

    Returns:
        str: The raw JSON response from the Wikipedia API, containing a list of
            search results.
    """
    import requests

    params = {"action": "query", "list": "search", "srsearch": query, "format": "json"}
    response = requests.get(
        "https://en.wikipedia.org/w/api.php",
        headers={
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
        },
        params=params,
    )
    return response.json()


def search_arxiv(query: str, num_results: int = 10) -> str:
    """
    Searches for academic papers and preprints on ArXiv.

    Use this tool when the user's query is scientific or technical in nature
    and they are likely looking for research papers, articles, or academic
    publications.

    Args:
        query (str): The search query, which can include keywords, author
            names, or titles.
        num_results (int, optional): The maximum number of results to return.
            Defaults to 10.

    Returns:
        str: The raw XML response from the ArXiv API, containing a list of
            matching papers.
    """
    import requests

    params = {"search_query": f"all:{query}", "start": 0, "max_results": num_results}
    response = requests.get(
        "http://export.arxiv.org/api/query",
        headers={
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
        },
        params=params,
    )
    return response.content


async def _fetch_page_content(url: str) -> tuple[str, list[str]]:
    """Fetches the HTML content and all absolute links from a URL."""
    user_agent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"
    try:
        from playwright.async_api import async_playwright

        async with async_playwright() as p:
            browser = await p.chromium.launch(headless=True)
            page = await browser.new_page()
            await page.set_extra_http_headers({"User-Agent": user_agent})
            try:
                await page.goto(url, wait_until="networkidle", timeout=30000)
                await page.wait_for_load_state("domcontentloaded")
                content = await page.content()
                links = await page.eval_on_selector_all(
                    "a[href]",
                    """
                    (elements, baseUrl) => elements.map(el => {
                        const href = el.getAttribute('href');
                        if (!href || href.startsWith('#')) return null;
                        try {
                            return new URL(href, baseUrl).href;
                        } catch (e) {
                            return null;
                        }
                    }).filter(href => href !== null)
                    """,
                    url,
                )
                return content, links
                # return json.dumps({"content": content, "links": links})
            finally:
                await browser.close()
    except Exception:
        import requests
        from bs4 import BeautifulSoup

        response = requests.get(url, headers={"User-Agent": user_agent})
        if response.status_code != 200:
            raise Exception(
                f"Unable to retrieve page content. Status code: {response.status_code}"
            )
        content = response.text
        soup = BeautifulSoup(content, "html.parser")
        links = [
            urljoin(url, a["href"])
            for a in soup.find_all("a", href=True)
            if not a["href"].startswith("#")
        ]
        return content, links
        # return json.dumps({"content": content, "links": links})


def _convert_html_to_markdown(html_text: str) -> str:
    """Converts HTML content to a clean Markdown representation."""
    from bs4 import BeautifulSoup
    from markdownify import markdownify as md

    soup = BeautifulSoup(html_text, "html.parser")
    # Remove non-content tags
    for tag in soup(
        ["script", "link", "meta", "style", "header", "footer", "nav", "aside"]
    ):
        tag.decompose()
    # Convert the cleaned HTML to Markdown
    return md(str(soup))



import os

from my_app_name._zrb.config import ACTIVATE_VENV_SCRIPT, APP_DIR
from my_app_name._zrb.entity.add_entity_util import (
    get_add_permission_migration_script,
    get_auth_migration_version_dir,
    get_existing_auth_migration_file_names,
    get_existing_auth_migration_xcom_key,
    get_remove_permission_migration_script,
    is_gateway_navigation_config_file,
    is_in_app_schema_dir,
    is_in_module_entity_dir,
    is_in_module_entity_test_dir,
    is_module_api_client_file,
    is_module_client_file,
    is_module_direct_client_file,
    is_module_gateway_subroute_file,
    is_module_gateway_subroute_view_file,
    is_module_migration_metadata_file,
    is_module_route_file,
    update_api_client_file,
    update_client_file,
    update_direct_client_file,
    update_gateway_navigation_config_file,
    update_gateway_subroute_file,
    update_migration_metadata_file,
    update_route_file,
)
from my_app_name._zrb.format_task import format_my_app_name_code
from my_app_name._zrb.group import app_create_group
from my_app_name._zrb.input import (
    existing_module_input,
    new_entity_column_input,
    new_entity_input,
    plural_entity_input,
)
from my_app_name._zrb.util import (
    cd_module_script,
    get_existing_module_names,
    get_existing_schema_names,
    set_create_migration_db_url_env,
    set_env,
)
from my_app_name._zrb.venv_task import prepare_venv

from zrb import (
    AnyContext,
    Cmd,
    CmdTask,
    ContentTransformer,
    EnvFile,
    Scaffolder,
    Task,
    Xcom,
    make_task,
)
from zrb.util.codemod.modify_function import replace_function_code
from zrb.util.file import read_file, write_file
from zrb.util.string.conversion import to_snake_case


@make_task(
    name="validate-add-my-app-name-entity",
    input=[
        existing_module_input,
        new_entity_input,
        plural_entity_input,
        new_entity_column_input,
    ],
    retries=0,
)
async def validate_add_my_app_name_entity(ctx: AnyContext):
    module_name = to_snake_case(ctx.input.module)
    if module_name not in get_existing_module_names():
        raise ValueError(f"Module not exist: {module_name}")
    schema_name = to_snake_case(ctx.input.entity)
    if schema_name in get_existing_schema_names():
        raise ValueError(f"Schema already exists: {schema_name}")


scaffold_my_app_name_entity = Scaffolder(
    name="scaffold-my-app-name-entity",
    input=[
        existing_module_input,
        new_entity_input,
        plural_entity_input,
        new_entity_column_input,
    ],
    source_path=os.path.join(os.path.dirname(__file__), "template", "app_template"),
    render_source_path=False,
    destination_path=APP_DIR,
    transform_path={
        "my_module": "{to_snake_case(ctx.input.module)}",
        "my-module": "{to_kebab_case(ctx.input.module)}",
        "my_entity": "{to_snake_case(ctx.input.entity)}",
        "my-entity": "{to_kebab_case(ctx.input.entity)}",
    },
    transform_content=[
        # Schema tranformation (my_app_name/schema/snake_entity_name)
        ContentTransformer(
            name="transform-schema-file",
            match=is_in_app_schema_dir,
            transform={
                "MyEntity": "{to_pascal_case(ctx.input.entity)}",
                "my_entities": "{to_snake_case(ctx.input.plural)}",
                "my_column": "{to_snake_case(ctx.input.column)}",
            },
        ),
        # Common module's entity transformation
        # (my_app_name/module/snake_module_name/service/snake_entity_name)
        ContentTransformer(
            name="transform-module-entity-dir",
            match=is_in_module_entity_dir,
            transform={
                "my_module": "{to_snake_case(ctx.input.module)}",
                "MyEntity": "{to_pascal_case(ctx.input.entity)}",
                "my_entity": "{to_snake_case(ctx.input.entity)}",
                "my_entities": "{to_snake_case(ctx.input.plural)}",
                "my-entities": "{to_kebab_case(ctx.input.plural)}",
            },
        ),
        # Test transformation (my_app_name/test/my_module/my_entity)
        ContentTransformer(
            name="transform-module-entity-test-dir",
            match=is_in_module_entity_test_dir,
            transform={
                "my_module": "{to_snake_case(ctx.input.module)}",
                "my_entity": "{to_snake_case(ctx.input.entity)}",
                "my-entity": "{to_kebab_case(ctx.input.entity)}",
                "my-entities": "{to_kebab_case(ctx.input.plural)}",
                "my_column": "{to_snake_case(ctx.input.column)}",
            },
        ),
        # Add entity to migration metadata
        # (my_app_name/module/snake_module_name/migration_metadata.py)
        ContentTransformer(
            name="transform-module-migration-metadata",
            match=is_module_migration_metadata_file,
            transform=update_migration_metadata_file,
        ),
        # Update API Client
        # (my_app_name/module/snake_module_name/client/snake_module_name_api_client.py)
        ContentTransformer(
            name="transform-module-api-client",
            match=is_module_api_client_file,
            transform=update_api_client_file,
        ),
        # Update Direct Client
        # (my_app_name/module/snake_module_name/client/snake_module_name_direct_client.py)
        ContentTransformer(
            name="transform-module-direct-client",
            match=is_module_direct_client_file,
            transform=update_direct_client_file,
        ),
        # Update Client
        # (my_app_name/module/snake_module_name/client/snake_module_name_client.py)
        ContentTransformer(
            name="transform-module-any-client",
            match=is_module_client_file,
            transform=update_client_file,
        ),
        # Update module route (my_app_name/module/route.py)
        ContentTransformer(
            name="transform-module-route",
            match=is_module_route_file,
            transform=update_route_file,
        ),
        # Update module gateway subroute
        # (my_app_name/module/gateway/subroute/snake_module_name.py)
        ContentTransformer(
            name="transform-module-gateway-subroute",
            match=is_module_gateway_subroute_file,
            transform=update_gateway_subroute_file,
        ),
        # Update module gateway subroute view
        # (my_app_name/module/gateway/view/content/kebab-module-name/kebab-entity-name.py)
        ContentTransformer(
            name="transform-module-gateway-subroute-view",
            match=is_module_gateway_subroute_view_file,
            transform={
                "My Entity": "{to_human_case(ctx.input.entity).title()}",
                "my-entities": "{to_kebab_case(ctx.input.plural)}",
                "My Column": "{to_human_case(ctx.input.column).title()}",
                "my_column": "{to_snake_case(ctx.input.column)}",
            },
        ),
        # Register entity's page to my_app_name/gateway/config/navigation.py
        ContentTransformer(
            name="transform-gateway-navigation-config",
            match=is_gateway_navigation_config_file,
            transform=update_gateway_navigation_config_file,
        ),
    ],
    retries=0,
    upstream=validate_add_my_app_name_entity,
)

create_my_app_name_entity_migration = CmdTask(
    name="create-my-app-name-entity-migration",
    input=[
        existing_module_input,
        new_entity_input,
    ],
    env=EnvFile(path=os.path.join(APP_DIR, "template.env")),
    cwd=APP_DIR,
    cmd=[
        ACTIVATE_VENV_SCRIPT,
        Cmd(lambda ctx: set_create_migration_db_url_env(ctx.input.module)),
        Cmd(lambda ctx: set_env("MY_APP_NAME_MODULES", ctx.input.module)),
        Cmd(lambda ctx: cd_module_script(ctx.input.module)),
        "alembic upgrade head",
        Cmd(
            'alembic revision --autogenerate -m "create_{to_snake_case(ctx.input.entity)}_table"',  # noqa
        ),
    ],
    render_cmd=False,
    retries=0,
    upstream=[
        prepare_venv,
        scaffold_my_app_name_entity,
    ],
)


@make_task(
    name="inspect-my-app-name-auth-migration",
    input=new_entity_input,
    retries=0,
    upstream=scaffold_my_app_name_entity,
)
def inspect_my_app_name_auth_migration(ctx: AnyContext):
    """Getting existing migration files in auth module"""
    migration_file_names = get_existing_auth_migration_file_names()
    xcom_key = get_existing_auth_migration_xcom_key(ctx)
    if xcom_key not in ctx.xcom:
        ctx.xcom[xcom_key] = Xcom([])
    ctx.xcom[xcom_key].push(migration_file_names)


create_my_app_name_entity_permission = CmdTask(
    name="create-my-app-name-entity-permission",
    input=[
        new_entity_input,
    ],
    env=EnvFile(path=os.path.join(APP_DIR, "template.env")),
    cwd=APP_DIR,
    cmd=[
        ACTIVATE_VENV_SCRIPT,
        Cmd(lambda ctx: set_create_migration_db_url_env("auth")),
        Cmd(lambda ctx: set_env("MY_APP_NAME_MODULES", "auth")),
        Cmd(lambda ctx: cd_module_script("auth")),
        "alembic upgrade head",
        Cmd(
            'alembic revision --autogenerate -m "create_{to_snake_case(ctx.input.entity)}_permission"',  # noqa
        ),
    ],
    render_cmd=False,
    retries=0,
    upstream=[
        prepare_venv,
        inspect_my_app_name_auth_migration,
    ],
)


@make_task(
    name="update-my-app-name-entity-permission",
    input=new_entity_input,
    retries=0,
    upstream=create_my_app_name_entity_permission,
)
def update_my_app_name_entity_permission(ctx: AnyContext):
    xcom_key = get_existing_auth_migration_xcom_key(ctx)
    existing_migration_file_names = ctx.xcom[xcom_key].pop()
    current_migration_file_names = get_existing_auth_migration_file_names()
    new_migration_file_names = [
        file_name
        for file_name in current_migration_file_names
        if file_name not in existing_migration_file_names
    ]
    if len(new_migration_file_names) == 0:
        raise Exception("No migration file created")
    new_migration_file_path = os.path.join(
        get_auth_migration_version_dir(), new_migration_file_names[0]
    )
    new_migration_code = read_file(
        new_migration_file_path,
        {
            "from alembic import op": "\n".join(
                [
                    "from alembic import op",
                    "from module.auth.migration_metadata import metadata",
                ]
            ),
        },
    )
    new_migration_code = replace_function_code(
        new_migration_code, "upgrade", get_add_permission_migration_script(ctx)
    )
    new_migration_code = replace_function_code(
        new_migration_code, "downgrade", get_remove_permission_migration_script(ctx)
    )
    write_file(new_migration_file_path, new_migration_code)


add_my_app_name_entity = app_create_group.add_task(
    Task(
        name="add-my-app-name-entity",
        description="📦 Create new entity on a module",
        upstream=[
            create_my_app_name_entity_migration,
            update_my_app_name_entity_permission,
        ],
        successor=format_my_app_name_code,
        retries=0,
    ),
    alias="entity",
)



import os
from typing import TYPE_CHECKING, Any, Callable

from zrb.config.config import CFG

if TYPE_CHECKING:
    from pydantic_ai.models import Model
    from pydantic_ai.providers import Provider
    from pydantic_ai.settings import ModelSettings


class LLMConfig:
    def __init__(
        self,
        default_model_name: str | None = None,
        default_base_url: str | None = None,
        default_api_key: str | None = None,
        default_persona: str | None = None,
        default_system_prompt: str | None = None,
        default_interactive_system_prompt: str | None = None,
        default_special_instruction_prompt: str | None = None,
        default_summarization_prompt: str | None = None,
        default_summarize_history: bool | None = None,
        default_history_summarization_token_threshold: int | None = None,
        default_modes: list[str] | None = None,
        default_model: "Model | None" = None,
        default_model_settings: "ModelSettings | None" = None,
        default_model_provider: "Provider | None" = None,
        default_yolo_mode: bool | None = None,
    ):
        self.__internal_default_prompt: dict[str, str] = {}
        self._default_model_name = default_model_name
        self._default_model_base_url = default_base_url
        self._default_model_api_key = default_api_key
        self._default_persona = default_persona
        self._default_system_prompt = default_system_prompt
        self._default_interactive_system_prompt = default_interactive_system_prompt
        self._default_special_instruction_prompt = default_special_instruction_prompt
        self._default_summarization_prompt = default_summarization_prompt
        self._default_summarize_history = default_summarize_history
        self._default_history_summarization_token_threshold = (
            default_history_summarization_token_threshold
        )
        self._default_modes = default_modes
        self._default_model = default_model
        self._default_model_settings = default_model_settings
        self._default_model_provider = default_model_provider
        self._default_yolo_mode = default_yolo_mode

    def _get_internal_default_prompt(self, name: str) -> str:
        if name not in self.__internal_default_prompt:
            file_path = os.path.join(
                os.path.dirname(__file__), "default_prompt", f"{name}.md"
            )
            with open(file_path, "r") as f:
                self.__internal_default_prompt[name] = f.read().strip()
        return self.__internal_default_prompt[name]

    def _get_property(
        self,
        instance_var: Any,
        config_var: Any,
        default_func: Callable[[], Any],
    ) -> Any:
        if instance_var is not None:
            return instance_var
        if config_var is not None:
            return config_var
        return default_func()

    @property
    def default_model_name(self) -> str | None:
        return self._get_property(self._default_model_name, CFG.LLM_MODEL, lambda: None)

    @property
    def default_model_base_url(self) -> str | None:
        return self._get_property(
            self._default_model_base_url, CFG.LLM_BASE_URL, lambda: None
        )

    @property
    def default_model_api_key(self) -> str | None:
        return self._get_property(
            self._default_model_api_key, CFG.LLM_API_KEY, lambda: None
        )

    @property
    def default_model_settings(self) -> "ModelSettings | None":
        return self._get_property(self._default_model_settings, None, lambda: None)

    @property
    def default_model_provider(self) -> "Provider | str":
        if self._default_model_provider is not None:
            return self._default_model_provider
        if self.default_model_base_url is None and self.default_model_api_key is None:
            return "openai"
        from pydantic_ai.providers.openai import OpenAIProvider

        return OpenAIProvider(
            base_url=self.default_model_base_url, api_key=self.default_model_api_key
        )

    @property
    def default_system_prompt(self) -> str:
        return self._get_property(
            self._default_system_prompt,
            CFG.LLM_SYSTEM_PROMPT,
            lambda: self._get_internal_default_prompt("system_prompt"),
        )

    @property
    def default_interactive_system_prompt(self) -> str:
        return self._get_property(
            self._default_interactive_system_prompt,
            CFG.LLM_INTERACTIVE_SYSTEM_PROMPT,
            lambda: self._get_internal_default_prompt("interactive_system_prompt"),
        )

    @property
    def default_persona(self) -> str:
        return self._get_property(
            self._default_persona,
            CFG.LLM_PERSONA,
            lambda: self._get_internal_default_prompt("persona"),
        )

    @property
    def default_modes(self) -> list[str]:
        return self._get_property(
            self._default_modes, CFG.LLM_MODES, lambda: ["coding"]
        )

    @property
    def default_special_instruction_prompt(self) -> str:
        return self._get_property(
            self._default_special_instruction_prompt,
            CFG.LLM_SPECIAL_INSTRUCTION_PROMPT,
            lambda: "",
        )

    @property
    def default_summarization_prompt(self) -> str:
        return self._get_property(
            self._default_summarization_prompt,
            CFG.LLM_SUMMARIZATION_PROMPT,
            lambda: self._get_internal_default_prompt("summarization_prompt"),
        )

    @property
    def default_model(self) -> "Model | str":
        if self._default_model is not None:
            return self._default_model
        model_name = self.default_model_name
        if model_name is None:
            return "openai:gpt-4o"
        from pydantic_ai.models.openai import OpenAIModel

        return OpenAIModel(
            model_name=model_name,
            provider=self.default_model_provider,
        )

    @property
    def default_summarize_history(self) -> bool:
        return self._get_property(
            self._default_summarize_history, CFG.LLM_SUMMARIZE_HISTORY, lambda: False
        )

    @property
    def default_history_summarization_token_threshold(self) -> int:
        return self._get_property(
            self._default_history_summarization_token_threshold,
            CFG.LLM_HISTORY_SUMMARIZATION_TOKEN_THRESHOLD,
            lambda: 1000,
        )

    @property
    def default_yolo_mode(self) -> bool:
        return self._get_property(
            self._default_yolo_mode, CFG.LLM_YOLO_MODE, lambda: False
        )

    def set_default_persona(self, persona: str):
        self._default_persona = persona

    def set_default_system_prompt(self, system_prompt: str):
        self._default_system_prompt = system_prompt

    def set_default_interactive_system_prompt(self, interactive_system_prompt: str):
        self._default_interactive_system_prompt = interactive_system_prompt

    def set_default_special_instruction_prompt(self, special_instruction_prompt: str):
        self._default_special_instruction_prompt = special_instruction_prompt

    def set_default_modes(self, modes: list[str]):
        self._default_modes = modes

    def add_default_mode(self, mode: str):
        if self._default_modes is None:
            self._default_modes = []
        self._default_modes.append(mode)

    def remove_default_mode(self, mode: str):
        if self._default_modes is None:
            self._default_modes = []
        self._default_modes.remove(mode)

    def set_default_summarization_prompt(self, summarization_prompt: str):
        self._default_summarization_prompt = summarization_prompt

    def set_default_model_name(self, model_name: str):
        self._default_model_name = model_name

    def set_default_model_api_key(self, model_api_key: str):
        self._default_model_api_key = model_api_key

    def set_default_model_base_url(self, model_base_url: str):
        self._default_model_base_url = model_base_url

    def set_default_model_provider(self, provider: "Provider | str"):
        self._default_model_provider = provider

    def set_default_model(self, model: "Model | str"):
        self._default_model = model

    def set_default_summarize_history(self, summarize_history: bool):
        self._default_summarize_history = summarize_history

    def set_default_history_summarization_token_threshold(
        self, history_summarization_token_threshold: int
    ):
        self._default_history_summarization_token_threshold = (
            history_summarization_token_threshold
        )

    def set_default_model_settings(self, model_settings: "ModelSettings"):
        self._default_model_settings = model_settings

    def set_default_yolo_mode(self, yolo_mode: bool):
        self._default_yolo_mode = yolo_mode


llm_config = LLMConfig()



import json
from typing import TYPE_CHECKING

from zrb.config.web_auth_config import WebAuthConfig
from zrb.group.any_group import AnyGroup
from zrb.runner.common_util import get_task_str_kwargs
from zrb.runner.web_util.user import get_user_from_request
from zrb.task.any_task import AnyTask
from zrb.util.group import NodeNotFoundError, extract_node_from_args

if TYPE_CHECKING:
    # We want fastapi to only be loaded when necessary to decrease footprint
    from fastapi import FastAPI


def serve_task_input_api(
    app: "FastAPI",
    root_group: AnyGroup,
    web_auth_config: WebAuthConfig,
) -> None:
    from fastapi import Query, Request
    from fastapi.responses import JSONResponse

    @app.get("/api/v1/task-inputs/{path:path}", response_model=dict[str, str])
    async def get_default_inputs_api(
        path: str,
        request: Request,
        query: str = Query("{}", description="JSON encoded inputs"),
    ) -> dict[str, str]:
        """
        Getting input completion for path
        """
        user = await get_user_from_request(web_auth_config, request)
        args = path.strip("/").split("/")
        try:
            task, _, _ = extract_node_from_args(root_group, args)
        except NodeNotFoundError:
            return JSONResponse(content={"detail": "Not found"}, status_code=404)
        if isinstance(task, AnyTask):
            if not user.can_access_task(task):
                return JSONResponse(content={"detail": "Forbidden"}, status_code=403)
            str_kwargs = json.loads(query)
            task_str_kwargs = get_task_str_kwargs(
                task=task, str_args=[], str_kwargs=str_kwargs, cli_mode=False
            )
            return task_str_kwargs
        return JSONResponse(content={"detail": "Not found"}, status_code=404)



import os
from functools import partial

from zrb.attr.type import BoolAttr, IntAttr, StrAttr
from zrb.cmd.cmd_result import CmdResult
from zrb.cmd.cmd_val import AnyCmdVal, CmdVal, SingleCmdVal
from zrb.config.config import CFG
from zrb.context.any_context import AnyContext
from zrb.env.any_env import AnyEnv
from zrb.input.any_input import AnyInput
from zrb.task.any_task import AnyTask
from zrb.task.base_task import BaseTask
from zrb.util.attr import get_int_attr, get_str_attr
from zrb.util.cmd.command import check_unrecommended_commands, run_command
from zrb.util.cmd.remote import get_remote_cmd_script
from zrb.xcom.xcom import Xcom


class CmdTask(BaseTask):
    def __init__(
        self,
        name: str,
        color: int | None = None,
        icon: str | None = None,
        description: str | None = None,
        cli_only: bool = False,
        input: list[AnyInput | None] | AnyInput | None = None,
        env: list[AnyEnv | None] | AnyEnv | None = None,
        shell: StrAttr | None = None,
        render_shell: bool = True,
        shell_flag: StrAttr | None = None,
        render_shell_flag: bool = True,
        remote_host: StrAttr | None = None,
        render_remote_host: bool = True,
        remote_port: IntAttr | None = None,
        render_remote_port: bool = True,
        remote_user: StrAttr | None = None,
        render_remote_user: bool = True,
        remote_password: StrAttr | None = None,
        render_remote_password: bool = True,
        remote_ssh_key: StrAttr | None = None,
        render_remote_ssh_key: bool = True,
        cmd: CmdVal = "",
        render_cmd: bool = True,
        cwd: str | None = None,
        render_cwd: bool = True,
        plain_print: bool = False,
        warn_unrecommended_command: bool | None = None,
        max_output_line: int = 1000,
        max_error_line: int = 1000,
        is_interactive: bool = False,
        execute_condition: BoolAttr = True,
        retries: int = 2,
        retry_period: float = 0,
        readiness_check: list[AnyTask] | AnyTask | None = None,
        readiness_check_delay: float = 0.5,
        readiness_check_period: float = 5,
        readiness_failure_threshold: int = 1,
        readiness_timeout: int = 60,
        monitor_readiness: bool = False,
        upstream: list[AnyTask] | AnyTask | None = None,
        fallback: list[AnyTask] | AnyTask | None = None,
        successor: list[AnyTask] | AnyTask | None = None,
    ):
        super().__init__(
            name=name,
            color=color,
            icon=icon,
            description=description,
            cli_only=cli_only,
            input=input,
            env=env,
            execute_condition=execute_condition,
            retries=retries,
            retry_period=retry_period,
            readiness_check=readiness_check,
            readiness_check_delay=readiness_check_delay,
            readiness_check_period=readiness_check_period,
            readiness_failure_threshold=readiness_failure_threshold,
            readiness_timeout=readiness_timeout,
            monitor_readiness=monitor_readiness,
            upstream=upstream,
            fallback=fallback,
            successor=successor,
        )
        self._shell = shell
        self._render_shell = render_shell
        self._shell_flag = shell_flag
        self._render_shell_flag = render_shell_flag
        self._remote_host = remote_host
        self._render_remote_host = render_remote_host
        self._remote_port = remote_port
        self._render_remote_port = render_remote_port
        self._remote_user = remote_user
        self._render_remote_user = render_remote_user
        self._remote_password = remote_password
        self._render_remote_password = render_remote_password
        self._remote_ssh_key = remote_ssh_key
        self._render_remote_ssh_key = render_remote_ssh_key
        self._cmd = cmd
        self._render_cmd = render_cmd
        self._cwd = cwd
        self._render_cwd = render_cwd
        self._max_output_line = max_output_line
        self._max_error_line = max_error_line
        self._should_plain_print = plain_print
        self._should_warn_unrecommended_command = warn_unrecommended_command
        self._is_interactive = is_interactive

    async def _exec_action(self, ctx: AnyContext) -> CmdResult:
        """Turn _cmd attribute into subprocess.Popen and execute it as task's action.

        Args:
            session (AnySession): The shared session.

        Returns:
            Any: The result of the action execution.
        """
        cmd_script = self._get_cmd_script(ctx)
        ctx.log_debug(f"Script: {self.__get_multiline_repr(cmd_script)}")
        shell = self._get_shell(ctx)
        ctx.log_debug(f"Shell: {shell}")
        shell_flag = self._get_shell_flag(ctx)
        cwd = self._get_cwd(ctx)
        ctx.log_debug(f"Working directory: {cwd}")
        env_map = self.__get_env_map(ctx)
        ctx.log_debug(f"Environment map: {env_map}")
        if self._get_should_warn_unrecommended_commands():
            self._check_unrecommended_commands(ctx, shell, cmd_script)
        ctx.log_info("Running script")
        print_method = (
            partial(ctx.print, plain=True) if self._should_plain_print else ctx.print
        )
        xcom_pid_key = f"{self.name}-pid"
        if xcom_pid_key not in ctx.xcom:
            ctx.xcom[xcom_pid_key] = Xcom([])
        cmd_result, return_code = await run_command(
            cmd=[shell, shell_flag, cmd_script],
            cwd=cwd,
            env_map=env_map,
            print_method=print_method,
            register_pid_method=lambda pid: ctx.xcom.get(xcom_pid_key).push(pid),
            max_output_line=self._max_output_line,
            max_error_line=self._max_error_line,
            is_interactive=self._is_interactive,
        )
        # Check for errors
        if return_code > 0:
            raise Exception(f"Process {self._name} exited ({return_code})")
        ctx.log_info(f"Exit status: {return_code}")
        return cmd_result

    def _get_should_warn_unrecommended_commands(self):
        if self._should_warn_unrecommended_command is None:
            return CFG.WARN_UNRECOMMENDED_COMMAND
        return self._should_warn_unrecommended_command

    def _check_unrecommended_commands(
        self, ctx: AnyContext, shell: str, cmd_script: str
    ):
        if shell.endswith("bash") or shell.endswith("zsh"):
            unrecommended_commands = check_unrecommended_commands(cmd_script)
            if unrecommended_commands:
                ctx.log_warning("The script contains unrecommended commands")
            for command, reason in unrecommended_commands.items():
                ctx.log_warning(f"- {command}: {reason}")

    def __get_env_map(self, ctx: AnyContext) -> dict[str, str]:
        envs = {key: val for key, val in ctx.env.items()}
        envs["_ZRB_SSH_PASSWORD"] = self._get_remote_password(ctx)
        envs["PYTHONBUFFERED"] = "1"
        return envs

    def _get_shell(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx, self._shell, CFG.DEFAULT_SHELL, auto_render=self._render_shell
        )

    def _get_shell_flag(self, ctx: AnyContext) -> str:
        default_shell_flags = {
            "node": "-e",
            "ruby": "-e",
            "php": "-r",
            "powershell": "/c",
        }
        default_shell_flag = default_shell_flags.get(self._get_shell(ctx).lower(), "-c")
        return get_str_attr(
            ctx,
            self._shell_flag,
            default_shell_flag,
            auto_render=self._render_shell_flag,
        )

    def _get_remote_host(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx, self._remote_host, "", auto_render=self._render_remote_host
        )

    def _get_remote_port(self, ctx: AnyContext) -> int:
        return get_int_attr(
            ctx, self._remote_port, 22, auto_render=self._render_remote_port
        )

    def _get_remote_user(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx, self._remote_user, "", auto_render=self._render_remote_user
        )

    def _get_remote_password(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx,
            self._remote_password,
            "",
            auto_render=self._render_remote_password,
        )

    def _get_remote_ssh_key(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx, self._remote_ssh_key, "", auto_render=self._render_remote_ssh_key
        )

    def _get_cwd(self, ctx: AnyContext) -> str:
        cwd = get_str_attr(ctx, self._cwd, os.getcwd(), auto_render=self._render_cwd)
        if cwd is None:
            cwd = os.getcwd()
        return os.path.abspath(cwd)

    def _get_cmd_script(self, ctx: AnyContext) -> str:
        if self._remote_host is None:
            return self._get_local_cmd_script(ctx)
        return self._get_remote_cmd_script(ctx)

    def _get_remote_cmd_script(self, ctx: AnyContext) -> str:
        return get_remote_cmd_script(
            cmd_script=self._get_local_cmd_script(ctx),
            host=self._get_remote_host(ctx),
            port=self._get_remote_port(ctx),
            user=self._get_remote_user(ctx),
            password="$_ZRB_SSH_PASSWORD",
            use_password=self._get_remote_password(ctx) != "",
            ssh_key=self._get_remote_ssh_key(ctx),
        )

    def _get_local_cmd_script(self, ctx: AnyContext) -> str:
        return self._render_cmd_val(ctx, self._cmd)

    def _render_cmd_val(self, ctx: AnyContext, cmd_val: CmdVal) -> str:
        if isinstance(cmd_val, list):
            cmd_val_list = [
                self.__render_single_cmd_val(ctx, single_cmd_val)
                for single_cmd_val in cmd_val
            ]
            return "\n".join(
                [cmd_val for cmd_val in cmd_val_list if cmd_val is not None]
            )
        return self.__render_single_cmd_val(ctx, cmd_val)

    def __render_single_cmd_val(
        self, ctx: AnyContext, single_cmd_val: SingleCmdVal
    ) -> str | None:
        if isinstance(single_cmd_val, AnyCmdVal):
            return single_cmd_val.to_str(ctx)
        if callable(single_cmd_val):
            return single_cmd_val(ctx)
        if isinstance(single_cmd_val, str):
            if self._render_cmd:
                return ctx.render(single_cmd_val)
            return single_cmd_val
        return None

    def __get_multiline_repr(self, text: str) -> str:
        lines_repr: list[str] = []
        lines = text.split("\n")
        if len(lines) == 1:
            return lines[0]
        for index, line in enumerate(lines):
            line_number_repr = str(index + 1).rjust(4, "0")
            lines_repr.append(f"   {line_number_repr} | {line}")
        return "\n" + "\n".join(lines_repr)



from fastapi.testclient import TestClient
from my_app_name.main import app
from my_app_name.test._util.access_token import get_admin_access_token


def test_delete_permission():
    client = TestClient(app, base_url="http://localhost")
    access_token = get_admin_access_token()
    new_permission_data = {
        "name": "to-be-deleted-permission",
        "description": "to-be-deleted-permission-description",
    }
    insert_response = client.post(
        "/api/v1/permissions",
        json=new_permission_data,
        headers={"Authorization": f"Bearer {access_token}"},
    )
    assert insert_response.status_code == 200
    id = insert_response.json().get("id")
    # deleting
    response = client.delete(
        f"/api/v1/permissions/{id}", headers={"Authorization": f"Bearer {access_token}"}
    )
    assert response.status_code == 200
    response_data = response.json()
    assert response_data.get("id") == id
    assert response_data.get("name") == "to-be-deleted-permission"
    assert response_data.get("description") == "to-be-deleted-permission-description"


def test_delete_permission_bulk():
    client = TestClient(app, base_url="http://localhost")
    access_token = get_admin_access_token()
    new_first_permission_data = {
        "name": "to-be-deleted-permission-bulk-1",
        "description": "to-be-deleted-permission-bulk-description-1",
    }
    new_second_permission_data = {
        "name": "to-be-deleted-permission-bulk-2",
        "description": "to-be-deleted-permission-bulk-description-2",
    }
    new_permission_data = [new_first_permission_data, new_second_permission_data]
    insert_response = client.post(
        "/api/v1/permissions/bulk",
        json=new_permission_data,
        headers={"Authorization": f"Bearer {access_token}"},
    )
    assert insert_response.status_code == 200
    ids = [row["id"] for row in insert_response.json()]
    # deleting (use client.request since client.delete doesn't support json param)
    response = client.request(
        "DELETE",
        f"/api/v1/permissions/bulk",
        json=ids,
        headers={"Authorization": f"Bearer {access_token}"},
    )
    assert response.status_code == 200
    response_data = response.json()
    # Data should match
    assert len([row["id"] for row in response_data if row["id"] in ids]) == 2
    assert new_first_permission_data["name"] in [row["name"] for row in response_data]
    assert new_second_permission_data["name"] in [row["name"] for row in response_data]
    assert new_first_permission_data["description"] in [
        row["description"] for row in response_data
    ]
    assert new_second_permission_data["description"] in [
        row["description"] for row in response_data
    ]
